"use strict";
// Copyright (c) Microsoft Corporation. All rights reserved. Licensed under the MIT license.
// See LICENSE in the project root for license information.
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const node_path_1 = __importDefault(require("node:path"));
const node_child_process_1 = require("node:child_process");
const node_core_library_1 = require("@rushstack/node-core-library");
const lookup_by_path_1 = require("@rushstack/lookup-by-path");
const heft_typescript_plugin_1 = require("@rushstack/heft-typescript-plugin");
const tapable_1 = require("tapable");
const TSC_TO_SWC_MODULE_MAP = {
    CommonJS: 'commonjs',
    ES2015: 'es6',
    ES2020: 'es6',
    ES2022: 'es6',
    ESNext: 'es6',
    Node16: 'nodenext',
    Node18: 'nodenext',
    NodeNext: 'nodenext',
    AMD: 'amd',
    None: undefined,
    UMD: 'umd',
    System: undefined,
    Preserve: undefined
};
const TSC_TO_SWC_TARGET_MAP = {
    ES2015: 'es2015',
    ES2016: 'es2016',
    ES2017: 'es2017',
    ES2018: 'es2018',
    ES2019: 'es2019',
    ES2020: 'es2020',
    ES2021: 'es2021',
    ES2022: 'es2022',
    ES2023: 'es2023',
    ES2024: 'es2024',
    ESNext: 'esnext',
    Latest: 'esnext',
    ES5: 'es5',
    ES3: 'es3',
    JSON: undefined
};
const PLUGIN_NAME = 'swc-isolated-transpile-plugin';
/**
 * @public
 */
class SwcIsolatedTranspilePlugin {
    constructor() {
        this.accessor = {
            hooks: {
                getSwcOptions: new tapable_1.SyncWaterfallHook(['swcOptions', 'format', 'target'])
            }
        };
    }
    apply(heftSession, heftConfiguration, pluginOptions = {}) {
        heftSession.hooks.run.tapPromise(PLUGIN_NAME, async () => {
            const { logger } = heftSession;
            await transpileProjectAsync(heftConfiguration, pluginOptions, logger, this.accessor);
        });
    }
}
exports.default = SwcIsolatedTranspilePlugin;
async function transpileProjectAsync(heftConfiguration, pluginOptions, logger, { hooks: { getSwcOptions: getSwcOptionsHook } }) {
    var _a;
    const { buildFolderPath } = heftConfiguration;
    const { emitKinds = [] } = pluginOptions;
    const { tool } = await (0, heft_typescript_plugin_1._loadTypeScriptToolAsync)({
        terminal: logger.terminal,
        heftConfiguration
    });
    const { ts } = tool;
    const tsconfigPath = (0, heft_typescript_plugin_1._getTsconfigFilePath)(heftConfiguration, pluginOptions.tsConfigPath);
    const parsedTsConfig = (0, heft_typescript_plugin_1._loadTsconfig)({ tool, tsconfigPath });
    if (!parsedTsConfig) {
        logger.terminal.writeLine('tsconfig.json not found. Skipping parse and transpile for this project.');
        return;
    }
    if (emitKinds.length < 1) {
        throw new Error('One or more emit kinds must be specified in the plugin options. To disable SWC transpilation, ' +
            'point "tsConfigPath" at a nonexistent file.');
    }
    logger.terminal.writeDebugLine('Loaded tsconfig', JSON.stringify(parsedTsConfig, undefined, 2));
    const { fileNames: filesFromTsConfig, options: tsConfigOptions } = parsedTsConfig;
    const { sourceMap, sourceRoot, experimentalDecorators, inlineSourceMap, useDefineForClassFields } = tsConfigOptions;
    const rootDirs = new Set(tsConfigOptions.rootDirs);
    if (tsConfigOptions.rootDir) {
        rootDirs.add(tsConfigOptions.rootDir);
    }
    const rootDirsPaths = new lookup_by_path_1.LookupByPath();
    for (const rootDir of rootDirs) {
        rootDirsPaths.setItem(rootDir, rootDir.length);
    }
    const sourceFilePaths = filesFromTsConfig.filter((filePath) => !filePath.endsWith('.d.ts'));
    logger.terminal.writeVerboseLine('Reading Config');
    const srcDir = node_core_library_1.Path.convertToSlashes(node_path_1.default.resolve(buildFolderPath, (_a = tsConfigOptions.rootDir) !== null && _a !== void 0 ? _a : 'src'));
    const sourceMaps = inlineSourceMap ? 'inline' : sourceMap;
    const externalSourceMaps = sourceMaps === true;
    function getOptionsByExtension({ formatOverride, targetOverride }) {
        var _a, _b, _c, _d, _e;
        const format = formatOverride !== undefined ? TSC_TO_SWC_MODULE_MAP[formatOverride] : undefined;
        if (format === undefined) {
            throw new Error(`Unsupported Module Kind: ${formatOverride && ts.ModuleKind[formatOverride]} for swc`);
        }
        logger.terminal.writeVerboseLine(`Transpiling to format: ${format}`);
        const target = targetOverride !== undefined ? TSC_TO_SWC_TARGET_MAP[targetOverride] : undefined;
        if (target === undefined) {
            throw new Error(`Unsupported Target: ${target && ts.ScriptTarget[target]} for swc`);
        }
        logger.terminal.writeVerboseLine(`Transpiling to target: ${target}`);
        const moduleConfig = {
            type: format,
            noInterop: tsConfigOptions.esModuleInterop === false
        };
        const parser = {
            syntax: 'typescript',
            decorators: experimentalDecorators,
            dynamicImport: true,
            tsx: false
        };
        // https://github.com/swc-project/swc-node/blob/e6cd8b83d1ce76a0abf770f52425704e5d2872c6/packages/register/read-default-tsconfig.ts#L131C7-L139C20
        const react = ((_c = (_b = (_a = tsConfigOptions.jsxFactory) !== null && _a !== void 0 ? _a : tsConfigOptions.jsxFragmentFactory) !== null && _b !== void 0 ? _b : tsConfigOptions.jsx) !== null && _c !== void 0 ? _c : tsConfigOptions.jsxImportSource)
            ? {
                pragma: tsConfigOptions.jsxFactory,
                pragmaFrag: tsConfigOptions.jsxFragmentFactory,
                importSource: (_d = tsConfigOptions.jsxImportSource) !== null && _d !== void 0 ? _d : 'react',
                runtime: ((_e = tsConfigOptions.jsx) !== null && _e !== void 0 ? _e : 0) >= ts.JsxEmit.ReactJSX ? 'automatic' : 'classic',
                useBuiltins: true
            }
            : undefined;
        let options = {
            cwd: buildFolderPath,
            root: srcDir,
            rootMode: 'root',
            configFile: false,
            swcrc: false,
            minify: false,
            sourceMaps: externalSourceMaps,
            inputSourceMap: true,
            sourceRoot,
            isModule: true,
            module: moduleConfig,
            jsc: {
                target,
                externalHelpers: tsConfigOptions.importHelpers,
                parser,
                transform: {
                    legacyDecorator: experimentalDecorators,
                    react,
                    useDefineForClassFields,
                    // This property is not included in the types, but is what makes swc-jest work
                    hidden: {
                        jest: format === 'commonjs'
                    }
                }
            }
        };
        if (getSwcOptionsHook.isUsed()) {
            options = getSwcOptionsHook.call(options, formatOverride, targetOverride);
        }
        logger.terminal.writeVerboseLine(`Transpile options: ${JSON.stringify(options, undefined, 2)}}`);
        logger.terminal.writeDebugLine(`Transpile options: ${options}`);
        const tsOptions = JSON.stringify(options);
        parser.tsx = true;
        const tsxOptions = JSON.stringify(options);
        return {
            ts: tsOptions,
            tsx: tsxOptions
        };
    }
    const outputOptions = new Map();
    for (const emitKind of emitKinds) {
        const { outDir } = emitKind;
        outputOptions.set(normalizeRelativeDir(outDir), getOptionsByExtension(emitKind));
    }
    const tasks = [];
    const requestMessage = {
        tasks,
        options: []
    };
    const indexForOptions = new Map();
    for (const srcFilePath of sourceFilePaths) {
        const rootPrefixLength = rootDirsPaths.findChildPath(srcFilePath);
        if (rootPrefixLength === undefined) {
            throw new Error(`Could not determine root prefix for ${srcFilePath}}`);
        }
        const relativeSrcFilePath = srcFilePath.slice(rootPrefixLength);
        const extensionIndex = relativeSrcFilePath.lastIndexOf('.');
        const tsx = endsWithCharacterX(relativeSrcFilePath);
        const relativeJsFilePath = `${relativeSrcFilePath.slice(0, extensionIndex)}.js`;
        for (const [outputPrefix, optionsByExtension] of outputOptions) {
            const jsFilePath = `${outputPrefix}${relativeJsFilePath}`;
            const mapFilePath = externalSourceMaps ? `${jsFilePath}.map` : undefined;
            const absoluteMapFilePath = `${buildFolderPath}/${mapFilePath}`;
            const relativeMapSrcFilePath = node_core_library_1.Path.convertToSlashes(node_path_1.default.relative(node_path_1.default.dirname(absoluteMapFilePath), srcFilePath));
            const options = tsx ? optionsByExtension.tsx : optionsByExtension.ts;
            let optionsIndex = indexForOptions.get(options);
            if (optionsIndex === undefined) {
                optionsIndex = requestMessage.options.push(options) - 1;
                indexForOptions.set(options, optionsIndex);
            }
            const item = {
                srcFilePath,
                relativeSrcFilePath: relativeMapSrcFilePath,
                optionsIndex,
                jsFilePath,
                mapFilePath
            };
            tasks.push(item);
        }
    }
    logger.terminal.writeLine(`Transpiling ${tasks.length} files...`);
    const result = await new Promise((resolve, reject) => {
        const workerPath = require.resolve('./TranspileWorker.js');
        const concurrency = Math.min(4, tasks.length, heftConfiguration.numberOfCores);
        // Due to https://github.com/rust-lang/rust/issues/91979 using worker_threads is not recommended for swc & napi-rs,
        // so we use child_process.fork instead.
        const childProcess = (0, node_child_process_1.fork)(workerPath, [buildFolderPath, `${concurrency}`]);
        childProcess.once('message', (message) => {
            // Shut down the worker.
            childProcess.send(false);
            // Node IPC messages are deserialized automatically.
            resolve(message);
        });
        childProcess.once('error', (error) => {
            reject(error);
        });
        childProcess.once('close', (closeExitCode, closeSignal) => {
            if (closeSignal) {
                reject(new Error(`Child process exited with signal: ${closeSignal}`));
            }
            else if (closeExitCode !== 0) {
                reject(new Error(`Child process exited with code: ${closeExitCode}`));
            }
        });
        childProcess.send(requestMessage);
    });
    const { errors, timings: transformTimes, durationMs } = result;
    printTiming(logger, transformTimes, 'Transformed');
    logger.terminal.writeLine(`Finished transpiling files in ${durationMs.toFixed(2)}ms`);
    const sortedErrors = errors.sort((x, y) => {
        const xPath = x[0];
        const yPath = y[0];
        return xPath > yPath ? 1 : xPath < yPath ? -1 : 0;
    });
    for (const [, error] of sortedErrors) {
        logger.emitError(new Error(error));
    }
}
function printTiming(logger, times, descriptor) {
    times.sort((x, y) => {
        return y[1] - x[1];
    });
    const timesCount = times.length;
    logger.terminal.writeVerboseLine(`${descriptor} ${timesCount} files at ${process.uptime()}`);
    if (timesCount > 0) {
        logger.terminal.writeVerboseLine(`Slowest files:`);
        for (let i = 0, len = Math.min(timesCount, 10); i < len; i++) {
            const [fileName, time] = times[i];
            logger.terminal.writeVerboseLine(`- ${fileName}: ${time.toFixed(2)}ms`);
        }
        const medianIndex = timesCount >> 1;
        const [medianFileName, medianTime] = times[medianIndex];
        logger.terminal.writeVerboseLine(`Median: (${medianFileName}): ${medianTime.toFixed(2)}ms`);
    }
}
function normalizeRelativeDir(relativeDir) {
    return relativeDir.startsWith('./')
        ? relativeDir.slice(2)
        : relativeDir.startsWith('/')
            ? relativeDir.slice(1)
            : relativeDir;
}
function endsWithCharacterX(filePath) {
    return filePath.charCodeAt(filePath.length - 1) === 120;
}
//# sourceMappingURL=SwcIsolatedTranspilePlugin.js.map